"""This module contains handle_aukus_post, which is the endpoint for AUKUS API requests"""

import copy
import os
from pathlib import Path
from typing import Optional

import requests
from fastapi import FastAPI, HTTPException
from fastapi.encoders import jsonable_encoder
from pydantic_settings import BaseSettings, SettingsConfigDict

from nrtk_jatic.api.aukus_schema import AukusDatasetSchema
from nrtk_jatic.api.schema import NrtkPerturbInputSchema


class Settings(BaseSettings):
    """Dataclass for NRTK API settings"""

    NRTK_IP: Optional[str] = None

    model_config = SettingsConfigDict(env_file=os.getcwd().split("nrtk-jatic")[0] + "nrtk-jatic/configs/AUKUS_app.env")


settings = Settings()
AUKUS_app = FastAPI()


def _check_input(data: AukusDatasetSchema) -> None:
    """Check input data and raise HTTPException if needed"""
    if data.data_format != "COCO":
        raise HTTPException(status_code=400, detail="Labels provided in incorrect format.")
    if not settings.NRTK_IP:
        raise HTTPException(status_code=400, detail="Provide NRTK_IP in AUKUS_app.env.")

    # Read NRTK configuration file and add relevant data to internalJSON
    if not os.path.isfile(data.nrtk_config):
        raise HTTPException(status_code=400, detail="Provided NRTK config is not a valid file.")


@AUKUS_app.post("/")
def handle_aukus_post(data: AukusDatasetSchema) -> list[AukusDatasetSchema]:
    """Format AUKUS request data to NRTK API format and return NRTK API data in AUKUS format"""
    _check_input(data)
    annotation_file = Path(data.uri) / data.labels[0]["iri"]

    nrtk_input = NrtkPerturbInputSchema(
        id=data.id,
        name=data.name,
        dataset_dir=data.uri,
        label_file=str(annotation_file),
        output_dir=data.output_dir,
        image_metadata=data.image_metadata,
        config_file=data.nrtk_config,
    )

    # Call 'handle_post' function with processed data and get the result
    out = requests.post(settings.NRTK_IP, json=jsonable_encoder(nrtk_input), timeout=3600).json()

    # Process the result and construct return JSONs
    return_jsons = []
    for i in range(len(out["datasets"])):
        dataset = out["datasets"][i]
        dataset_json = copy.deepcopy(data)
        dataset_json.uri = dataset["root_dir"]
        if dataset_json.labels:
            dataset_json.labels = [
                {
                    "name": f"{dataset_json.labels[0]['name']}_pertubation_{i}",
                    "iri": dataset["label_file"],
                    "objectCount": dataset_json.labels[0]["objectCount"],
                },
            ]
        return_jsons.append(dataset_json)

    return return_jsons



"""This module contains wrappers for NRTK perturbers for image classification"""

import copy
from collections.abc import Sequence
from typing import Optional

import numpy as np
from maite.protocols import ArrayLike
from maite.protocols.image_classification import (
    Augmentation,
    DatumMetadataBatchType,
    InputBatchType,
    TargetBatchType,
)
from nrtk.interfaces.image_metric import ImageMetric
from nrtk.interfaces.perturb_image import PerturbImage

IMG_CLASSIFICATION_BATCH_T = tuple[InputBatchType, TargetBatchType, DatumMetadataBatchType]


class JATICClassificationAugmentation(Augmentation):
    """Implementation of JATIC Augmentation for NRTK perturbers.

    Implementation of JATIC Augmentation for NRTK perturbers operating on a MAITE-protocol
    compliant Image Classification dataset.

    Parameters
    ----------
    augment : PerturbImage
        Augmentations to apply to an image.
    """

    def __init__(self, augment: PerturbImage) -> None:
        """Initialize augmentation wrapper"""
        self.augment = augment

    def __call__(self, batch: IMG_CLASSIFICATION_BATCH_T) -> IMG_CLASSIFICATION_BATCH_T:
        """Apply augmentations to the given data batch."""
        imgs, anns, metadata = batch

        # iterate over (parallel) elements in batch
        aug_imgs = list()  # list of individual augmented inputs
        aug_labels = list()  # list of individual class labels
        aug_metadata = list()  # list of individual image-level metadata

        for img, ann, md in zip(imgs, anns, metadata):
            # Perform augmentation
            aug_img = copy.deepcopy(img)
            aug_img = self.augment(np.asarray(aug_img), md)
            aug_height, aug_width = aug_img.shape[0:2]
            aug_imgs.append(aug_img)

            y_aug_labels = copy.deepcopy(ann)
            aug_labels.append(y_aug_labels)

            m_aug = copy.deepcopy(md)
            m_aug.update(
                {
                    "nrtk::perturber": self.augment.get_config(),
                    "image_info": {"width": aug_width, "height": aug_height},
                },
            )
            aug_metadata.append(m_aug)

        # return batch of augmented inputs, class labels and updated metadata
        return aug_imgs, aug_labels, aug_metadata


class JATICClassificationAugmentationWithMetric(Augmentation):
    """Implementation of JATIC augmentation wrapper for NRTK's Image metrics.

    Implementation of JATIC augmentation for NRTK metrics operating on a MAITE-protocol
    compliant image classification dataset.

    Parameters
    ----------
    augmentations : Optional[Sequence[Augmentation]]
        Optional task-specific sequence of JATIC augmentations to be applied on a given batch.
    metric : ImageMetric
        Image metric to be applied for a given image.
    """

    def __init__(self, augmentations: Optional[Sequence[Augmentation]], metric: ImageMetric) -> None:
        """Initialize augmentation with metric wrapper"""
        self.augmentations = augmentations
        self.metric = metric

    def __call__(self, batch: IMG_CLASSIFICATION_BATCH_T) -> IMG_CLASSIFICATION_BATCH_T:
        """Compute a specified image metric on the given batch."""
        imgs, anns, metadata = batch
        metric_aug_metadata = list()  # list of individual image-level metric metadata

        aug_imgs: Sequence[Optional[ArrayLike]] = list()
        if self.augmentations:
            aug_batch = batch
            for aug in self.augmentations:
                aug_batch = aug(aug_batch)
            aug_imgs, aug_anns, aug_metadata = aug_batch
        else:
            aug_imgs, aug_anns, aug_metadata = [None] * len(imgs), anns, metadata

        for img, aug_img, aug_md in zip(imgs, aug_imgs, aug_metadata):
            # Convert from channels-first to channels-last
            img_1 = np.transpose(img, (1, 2, 0))
            img_2 = None if aug_img is None else np.transpose(aug_img, (1, 2, 0))

            # Compute Image metric values
            metric_value = self.metric(img_1=img_1, img_2=img_2, additional_params=aug_md)
            metric_aug_md = copy.deepcopy(aug_md)
            metric_name = self.metric.__class__.__name__
            metric_aug_md.update({"nrtk::" + metric_name: metric_value})
            metric_aug_metadata.append(metric_aug_md)

        # return batch of augmented/original images, annotations and metric-updated metadata
        if self.augmentations:
            # type ignore was included to handle the dual Sequence[ArrrayLike] | List[None]
            # case for the augmented images.
            return aug_imgs, aug_anns, metric_aug_metadata  # type: ignore
        return imgs, aug_anns, metric_aug_metadata



"""This module contains wrappers for NRTK perturbers for object detection"""

import copy
from collections.abc import Sequence
from typing import Optional

import numpy as np
from maite.protocols import ArrayLike
from maite.protocols.object_detection import (
    Augmentation,
    DatumMetadataBatchType,
    InputBatchType,
    TargetBatchType,
)
from nrtk.interfaces.image_metric import ImageMetric
from nrtk.interfaces.perturb_image import PerturbImage

from nrtk_jatic.interop.object_detection.dataset import JATICDetectionTarget

OBJ_DETECTION_BATCH_T = tuple[InputBatchType, TargetBatchType, DatumMetadataBatchType]


class JATICDetectionAugmentation(Augmentation):
    """Implementation of JATIC Augmentation for NRTK perturbers.

    Implementation of JATIC Augmentation for NRTK perturbers
    operating on a MAITE-protocol compliant Object Detection dataset.

    Given a set of ground truth labels alongside an image and image
    metadata, JATICDetectionAugmentation will properly scale the
    labels in accordance with the change in the image scale due to
    applied pertubation. At this time JATICDetectionAugmentation does
    not support any other augmentation to the labels such as cropping,
    translation, or rotation.

    Parameters
    ----------
    augment : PerturbImage
        Augmentations to apply to an image.
    """

    def __init__(self, augment: PerturbImage) -> None:
        """Initialize augmentation wrapper"""
        self.augment = augment

    def __call__(self, batch: OBJ_DETECTION_BATCH_T) -> OBJ_DETECTION_BATCH_T:
        """Apply augmentations to the given data batch."""
        imgs, anns, metadata = batch

        # iterate over (parallel) elements in batch
        aug_imgs = list()  # list of individual augmented inputs
        aug_dets = list()  # list of individual object detection targets
        aug_metadata = list()  # list of individual image-level metadata

        for img, ann, md in zip(imgs, anns, metadata):
            # Perform augmentation
            aug_img = copy.deepcopy(img)
            aug_img = np.transpose(aug_img, (1, 2, 0))
            height, width = aug_img.shape[0:2]  # type: ignore
            aug_img = self.augment(np.asarray(aug_img), md)
            aug_height, aug_width = aug_img.shape[0:2]
            aug_imgs.append(np.transpose(aug_img, (2, 0, 1)))

            # Resize bounding boxes
            y_aug_boxes = copy.deepcopy(np.asarray(ann.boxes))
            y_aug_labels = copy.deepcopy(np.asarray(ann.labels))
            y_aug_scores = copy.deepcopy(np.asarray(ann.scores))
            y_aug_boxes[:, 0] *= aug_width / width
            y_aug_boxes[:, 1] *= aug_height / height
            y_aug_boxes[:, 2] *= aug_width / width
            y_aug_boxes[:, 3] *= aug_height / height
            aug_dets.append(JATICDetectionTarget(y_aug_boxes, y_aug_labels, y_aug_scores))

            m_aug = copy.deepcopy(md)
            m_aug.update({"nrtk::perturber": self.augment.get_config()})
            aug_metadata.append(m_aug)

        # return batch of augmented inputs, resized bounding boxes and updated metadata
        return aug_imgs, aug_dets, aug_metadata


class JATICDetectionAugmentationWithMetric(Augmentation):
    """Implementation of JATIC augmentation wrapper for NRTK's Image metrics.

    Implementation of JATIC augmentation for NRTK metrics operating on a MAITE-protocol
    compliant object detection dataset.

    Parameters
    ----------
    augmentations : Optional[Sequence[Augmentation]]
        Optional task-specific sequence of JATIC augmentations to be applied on a given batch.
    metric : ImageMetric
        Image metric to be applied for a given image.
    """

    def __init__(self, augmentations: Optional[Sequence[Augmentation]], metric: ImageMetric) -> None:
        """Initialize augmentation with metric wrapper"""
        self.augmentations = augmentations
        self.metric = metric

    def __call__(self, batch: OBJ_DETECTION_BATCH_T) -> OBJ_DETECTION_BATCH_T:
        """Compute a specified image metric on the given batch."""
        imgs, dets, metadata = batch
        metric_aug_metadata = list()  # list of individual image-level metric metadata

        aug_imgs: Sequence[Optional[ArrayLike]] = list()
        if self.augmentations:
            aug_batch = batch
            for aug in self.augmentations:
                aug_batch = aug(aug_batch)
            aug_imgs, aug_dets, aug_metadata = aug_batch
        else:
            aug_imgs, aug_dets, aug_metadata = [None] * len(imgs), dets, metadata

        for img, aug_img, aug_md in zip(imgs, aug_imgs, aug_metadata):
            # Convert from channels-first to channels-last
            img_1 = img
            img_2 = None if aug_img is None else aug_img

            # Compute Image metric values
            metric_value = self.metric(img_1=img_1, img_2=img_2, additional_params=aug_md)  # type: ignore
            metric_aug_md = copy.deepcopy(aug_md)
            metric_name = self.metric.__class__.__name__
            metric_aug_md.update({"nrtk::" + metric_name: metric_value})
            metric_aug_metadata.append(metric_aug_md)

        # return batch of augmented/original images, detections and metric-updated metadata
        if self.augmentations:
            # type ignore was included to handle the dual Sequence[ArrrayLike] | List[None]
            # case for the augmented images.
            return aug_imgs, aug_dets, metric_aug_metadata  # type: ignore
        return imgs, aug_dets, metric_aug_metadata



"""
This module contains wrappers for converting a COCO dataset or
a generic dataset to a MAITE dataset for object detection
"""

import copy
import logging
from collections.abc import Sequence
from dataclasses import dataclass
from pathlib import Path
from typing import Any

import numpy as np
from maite.protocols.object_detection import (
    Dataset,
    DatumMetadataType,
    InputType,
    TargetType,
)
from PIL import Image  # type: ignore

try:
    import kwcoco  # type: ignore

    is_usable = True
except ImportError:
    is_usable = False

OBJ_DETECTION_DATUM_T = tuple[InputType, TargetType, DatumMetadataType]

LOG = logging.getLogger(__name__)


@dataclass
class JATICDetectionTarget:
    """Dataclass for the datum-level JATIC output detection format."""

    boxes: np.ndarray
    labels: np.ndarray
    scores: np.ndarray


def _xyxy_bbox_xform(x: int, y: int, w: int, h: int) -> tuple[int, int, int, int]:
    return x, y, x + w, y + h


def _coco_to_maite_detections(coco_annotation: list) -> TargetType:
    num_anns = len(coco_annotation)
    boxes = np.zeros((num_anns, 4))
    for i, anns in enumerate(coco_annotation):
        box = list(map(int, anns["bbox"]))
        # convert box from xywh in xyxy format
        boxes[i, :] = np.asarray(_xyxy_bbox_xform(x=box[0], y=box[1], w=box[2], h=box[3]))

    labels = np.stack([int(anns["category_id"]) for anns in coco_annotation])
    scores = np.ones(num_anns)

    return JATICDetectionTarget(boxes, labels, scores)


if not is_usable:
    LOG.warning("COCOJATICObjectDetectionDataset requires additional dependencies, please install 'nrtk-jatic[tools]'")
else:

    class COCOJATICObjectDetectionDataset(Dataset):
        """Dataset class to convert a COCO dataset to a dataset compliant with JATIC's Object Detection protocol.

        Parameters
        ----------
        root : str
            The root directory of the dataset.
        kwcoco_dataset : kwcoco.CocoDataset
            The kwcoco COCODataset object.
        image_metadata : list[dict[str, Any]]
            A list of per-image metadata. Any metadata required by a perturber should be provided.
        """

        def __init__(
            self,
            root: str,
            kwcoco_dataset: kwcoco.CocoDataset,
            image_metadata: list[dict[str, Any]],
        ) -> None:
            """Initialize MAITE-compliant dataset from a COCO dataset"""
            self._root: Path = Path(root)
            image_dir = self._root / "images"
            self.all_img_paths = [image_dir / val["file_name"] for key, val in kwcoco_dataset.imgs.items()]
            self.all_image_ids = sorted({p.stem for p in self.all_img_paths})

            # Get all image filenames from the kwcoco object
            anns_image_ids = [
                {"coco_image_id": val["id"], "filename": val["file_name"]} for key, val in kwcoco_dataset.imgs.items()
            ]
            anns_image_ids = sorted(anns_image_ids, key=lambda d: d["filename"])

            # store sorted image paths
            self._images = sorted([p for p in self.all_img_paths if p.stem in self.all_image_ids])

            self._annotations = {}
            for image_id, anns_img_id in zip(self.all_image_ids, anns_image_ids):
                image_annotations = [
                    sub for sub in list(kwcoco_dataset.anns.values()) if sub["image_id"] == anns_img_id["coco_image_id"]
                ]
                # Convert annotations to maite detections format
                self._annotations[image_id] = _coco_to_maite_detections(image_annotations)

            self.classes = list(kwcoco_dataset.cats.values())

            self._image_metadata = copy.deepcopy(image_metadata)
            if len(self._image_metadata) != len(self.all_img_paths):
                raise ValueError("Image metadata length mismatch, metadata needed for every image")

        def __len__(self) -> int:
            """Returns the number of images in the dataset."""
            return len(self._images)

        def __getitem__(self, index: int) -> OBJ_DETECTION_DATUM_T:
            """Returns the dataset object at the given index."""
            image_path = self._images[index]
            image = Image.open(image_path)
            image_id = image_path.stem
            width, height = image.size
            annotation = self._annotations[image_id].labels
            num_objects = np.asarray(annotation).shape[0]
            uniq_objects = np.unique(annotation)
            num_unique_classes = uniq_objects.shape[0]
            unique_classes = [self.classes[int(idx)]["name"] for idx in uniq_objects.tolist()]

            self._image_metadata[index].update(
                dict(
                    id=image_id,
                    image_info=dict(
                        width=width,
                        height=height,
                        num_objects=num_objects,
                        num_unique_classes=num_unique_classes,
                        unique_classes=unique_classes,
                    ),
                ),
            )

            input_img, dets, metadata = (
                np.transpose(np.asarray(image), (2, 0, 1)),
                self._annotations[image_id],
                self._image_metadata[index],
            )

            return input_img, dets, metadata

        def get_img_path_list(self) -> list[Path]:
            """Returns the sorted list of absolute paths for the input images."""
            return sorted(self.all_img_paths)

        def get_categories(self) -> list[dict[str, Any]]:
            """Returns the list of categories for this dataset."""
            return self.classes


class JATICObjectDetectionDataset(Dataset):
    """Implementation of the JATIC Object Detection dataset wrapper for dataset images of varying sizes.

    Parameters
    ----------
    imgs : Sequence[np.ndarray]
        Sequence of images.
    dets : Sequence[ObjectDetectionTarget]
        Sequence of detections for each image.
    metadata : Sequence[dict[str, Any]]
        Sequence of custom metadata for each image.
    """

    def __init__(
        self,
        imgs: Sequence[np.ndarray],
        dets: Sequence[TargetType],
        metadata: Sequence[DatumMetadataType],
    ) -> None:
        """Initialize MAITE-compliant dataset"""
        self.imgs = imgs
        self.dets = dets
        self.metadata = metadata

    def __len__(self) -> int:
        """Returns the number of images in the dataset."""
        return len(self.imgs)

    def __getitem__(self, index: int) -> OBJ_DETECTION_DATUM_T:
        """Returns the dataset object at the given index."""
        return self.imgs[index], self.dets[index], self.metadata[index]



"""This module contains wrappers for converting a generic dataset to a MAITE dataset for image classification"""

from collections.abc import Sequence

import numpy as np
from maite.protocols.image_classification import (
    Dataset,
    DatumMetadataType,
    InputType,
    TargetType,
)

IMG_CLASSIFICATION_DATUM_T = tuple[InputType, TargetType, DatumMetadataType]


class JATICImageClassificationDataset(Dataset):
    """Implementation of the JATIC Image Classification dataset wrapper for dataset images of varying sizes.

    Parameters
    ----------
    imgs : Sequence[np.ndarray]
        Sequence of images.
    labels : Sequence[ArrayLike]
        Sequence of labels for each image.
    metadata : Sequence[Dict[str, Any]]
        Sequence of custom metadata for each image.
    """

    def __init__(
        self,
        imgs: Sequence[np.ndarray],
        labels: Sequence[TargetType],
        metadata: Sequence[DatumMetadataType],
    ) -> None:
        """Initialize MAITE-compliant dataset"""
        self.imgs = imgs
        self.labels = labels
        self.metadata = metadata

    def __len__(self) -> int:
        """Returns the number of images in the dataset."""
        return len(self.imgs)

    def __getitem__(self, index: int) -> IMG_CLASSIFICATION_DATUM_T:
        """Returns the dataset object at the given index."""
        return self.imgs[index], self.labels[index], self.metadata[index]



"""Define the nrtk_jatic package"""

import pkg_resources

# It is known that this will fail if package is not "installed" in the
# current environment. Additional support is pending defined use-case-driven
# requirements.
__version__ = pkg_resources.get_distribution(__name__).version



"""This module contains schemas for AUKUS API"""

from typing import Any, Optional

from pydantic import BaseModel


class AukusdataCollectionSchema(BaseModel):
    """Collection schema for AUKUS API"""

    # header params
    doc_type: str
    doc_version: str
    ism: dict[str, Any]
    last_update_time: str
    id: str
    name: str
    uri: str

    # Required Data Collection Params
    size: int
    description: str

    # Optional Data Collection Params
    local_region: Optional[str] = None
    collection_date_time: Optional[str] = None
    data_entries: Optional[int] = None
    source: Optional[dict[str, str]] = None
    data_formats: Optional[list[dict[str, Any]]] = None


class AukusDatasetSchema(BaseModel):
    """Dataset schema for AUKUS API"""

    # header params
    doc_type: str
    doc_version: str
    ism: dict[str, Any]
    last_update_time: str
    id: str
    name: str
    uri: str

    # Required Dataset Params
    size: str
    description: str
    data_collections: list[AukusdataCollectionSchema]
    data_format: str
    labels: list[dict[str, Any]]

    # NRTk specific param
    nrtk_config: str
    image_metadata: list[dict[str, Any]]
    output_dir: str

    # Optional Dataset Params
    tags: Optional[list[str]] = None



"""This module contains handle_post, which is the endpoint for NRTK API requests"""

from pathlib import Path

from fastapi import FastAPI, HTTPException

from nrtk_jatic.api.converters import build_factory
from nrtk_jatic.api.schema import (
    DatasetSchema,
    NrtkPerturbInputSchema,
    NrtkPerturbOutputSchema,
)
from nrtk_jatic.interop.object_detection.utils import dataset_to_coco
from nrtk_jatic.utils.nrtk_perturber import nrtk_perturber

try:
    from nrtk_jatic.api.converters import load_COCOJATIC_dataset

    is_usable = True
except ImportError:
    is_usable = False


app = FastAPI()


# Define a route for handling POST requests
@app.post("/")
def handle_post(data: NrtkPerturbInputSchema) -> NrtkPerturbOutputSchema:
    """Returns a collection of augmented datasets based parameters in data.

    :param data: NrtkPybsmPerturbInputSchema from schema.py

    :returns: NrtkPybsmPerturberOutputSchema from schema.py

    :raises: HTTPException upon failure
    """
    try:
        # Build pybsm factory
        perturber_factory = build_factory(data)

        # Load dataset
        if not is_usable:
            raise ImportError("This tool requires additional dependencies, please install `nrtk-jatic[tools]`")
        input_dataset = load_COCOJATIC_dataset(data)

        # Call nrtk_perturber
        augmented_datasets = nrtk_perturber(maite_dataset=input_dataset, perturber_factory=perturber_factory)

        # Format output
        datasets_out = list()
        img_filenames = [Path("images") / img_path.name for img_path in input_dataset.get_img_path_list()]
        for perturb_params, aug_dataset in augmented_datasets:
            full_output_dir = Path(data.output_dir) / perturb_params
            dataset_to_coco(
                dataset=aug_dataset,
                output_dir=full_output_dir,
                img_filenames=img_filenames,
                dataset_categories=input_dataset.get_categories(),
            )
            datasets_out.append(
                DatasetSchema(
                    root_dir=str(full_output_dir),
                    label_file="annotations.json",
                    metadata_file="image_metadata.json",
                ),
            )

        return NrtkPerturbOutputSchema(
            message="Data received successfully",
            datasets=datasets_out,
        )
    except Exception as e:
        raise HTTPException(status_code=400, detail=str(e)) from e



"""This module contains functions to convert input schema to NRTK objects"""

import json
import logging
import os

from nrtk.interfaces.perturb_image_factory import PerturbImageFactory
from smqtk_core.configuration import from_config_dict

from nrtk_jatic.api.schema import NrtkPerturbInputSchema

try:
    import kwcoco  # type: ignore

    from nrtk_jatic.interop.object_detection.dataset import (
        COCOJATICObjectDetectionDataset,
    )

    is_usable = True
except ImportError:
    is_usable = False

LOG = logging.getLogger(__name__)


def build_factory(data: NrtkPerturbInputSchema) -> PerturbImageFactory:
    """Returns a PerturbImageFactory based on scenario and sensor parameters in data.

    :param data: dictionary of Schema from schema.py
    """
    if not os.path.isfile(data.config_file):
        raise FileNotFoundError(f"Config file at {data.config_file} was not found")
    with open(data.config_file) as config_file:
        config = json.load(config_file)
        if "PerturberFactory" not in config:
            raise ValueError(f'Config file at {data.config_file} does not have "PerturberFactory" key')
        return from_config_dict(config["PerturberFactory"], PerturbImageFactory.get_impls())


if not is_usable:
    LOG.warning("load_COCOJATIC_dataset requires additional dependencies, please install 'nrtk-jatic[tools]'")
else:

    def load_COCOJATIC_dataset(  # noqa: N802
        data: NrtkPerturbInputSchema,
    ) -> COCOJATICObjectDetectionDataset:
        """Returns a COCOJATICObjectDetectionDataset based on dataset parameters in data.

        :param data: dictionary of Schema from schema.py
        """
        if not is_usable:
            raise ImportError("This tool requires additional dependencies, please install `nrtk-jatic[tools]`")
        kwcoco_dataset = kwcoco.CocoDataset(data.label_file)

        return COCOJATICObjectDetectionDataset(
            root=data.dataset_dir,
            kwcoco_dataset=kwcoco_dataset,
            image_metadata=data.image_metadata,
        )



"""This module contains dataset_to_coco, which converts a MAITE dataset to a COCO dataset"""

import json
import logging
from pathlib import Path
from typing import Any

import numpy as np
from maite.protocols.object_detection import Dataset
from PIL import Image  # type: ignore

try:
    import kwcoco  # type: ignore

    is_usable = True
except ImportError:
    is_usable = False


def _xywh_bbox_xform(x1: int, y1: int, x2: int, y2: int) -> tuple[int, int, int, int]:
    return x1, y1, x2 - x1, y2 - y1


def _create_annotations(dataset_categories: list[dict[str, Any]]) -> kwcoco.CocoDataset:
    annotations = kwcoco.CocoDataset()
    for cat in dataset_categories:
        annotations.add_category(name=cat["name"], supercategory=cat["supercategory"], id=cat["id"])
    return annotations


def dataset_to_coco(
    dataset: Dataset,
    output_dir: Path,
    img_filenames: list[Path],
    dataset_categories: list[dict[str, Any]],
) -> None:
    """Save dataset object to file as a COCO formatted dataset.

    :param dataset: MAITE-compliant object detection dataset
    :param output_dir: The location where data will be saved.
    :param img_filenames: Filenames of images to be saved.
    :param dataset_categories: A list of the categories related to this dataset.
        Each dictionary should contain the following keys: id, name, supercategory.
    """
    if len(img_filenames) != len(dataset):
        raise ValueError(f"Image filename and dataset length mismatch ({len(img_filenames)} != {len(dataset)})")
    if not is_usable:
        raise ImportError("This tool requires additional dependencies, please install `nrtk-jatic[tools]`")
    mod_metadata = list()

    annotations = _create_annotations(dataset_categories)

    for i in range(len(dataset)):
        image, dets, metadata = dataset[i]
        filename = output_dir / img_filenames[i]
        filename.parent.mkdir(parents=True, exist_ok=True)

        im = Image.fromarray(np.transpose(np.asarray(image), (1, 2, 0)))
        im.save(filename)

        labels = np.asarray(dets.labels)
        bboxes = np.asarray(dets.boxes)
        annotations.add_images([{"id": i, "file_name": str(filename)}])
        for lbl, bbox in zip(labels, bboxes):
            annotations.add_annotation(
                image_id=i,
                category_id=int(lbl),
                bbox=list(
                    _xywh_bbox_xform(
                        x1=int(bbox[0]),
                        y1=int(bbox[1]),
                        x2=int(bbox[2]),
                        y2=int(bbox[3]),
                    ),
                ),
            )

        mod_metadata.append(metadata)
    logging.info(f"Saved perturbed images to {output_dir}")

    metadata_file = output_dir / "image_metadata.json"

    with open(metadata_file, "w") as f:
        json.dump(mod_metadata, f)
    logging.info(f"Saved image metadata to {metadata_file}")

    annotations_file = output_dir / "annotations.json"
    annotations.dump(annotations_file)
    logging.info(f"Saved annotations to {annotations_file}")



"""This module contains nrtk_perturber_cli, which is a CLI script for running nrtk_perturber"""

import json
import logging
from pathlib import Path
from typing import Any, TextIO

import click  # type: ignore
from nrtk.interfaces.perturb_image_factory import PerturbImageFactory
from smqtk_core.configuration import from_config_dict, make_default_config

from nrtk_jatic.utils.nrtk_perturber import nrtk_perturber

try:
    import kwcoco  # type: ignore

    from nrtk_jatic.interop.object_detection.dataset import (
        COCOJATICObjectDetectionDataset,
    )
    from nrtk_jatic.interop.object_detection.utils import dataset_to_coco

    is_usable = True
except ImportError:
    is_usable = False


def _load_metadata(dataset_dir: str, kwcoco_dataset: kwcoco.CocoDataset) -> list[dict[str, Any]]:
    metadata_file = Path(dataset_dir) / "image_metadata.json"
    if not metadata_file.is_file():
        logging.warn(
            "Could not identify metadata file, assuming no metadata. Expected at '[dataset_dir]/image_metadata.json'",
        )
        return [dict() for _ in range(len(kwcoco_dataset.imgs))]
    logging.info(f"Loading metadata from {metadata_file}")
    with open(metadata_file) as f:
        return json.load(f)


def _set_logging(verbose: bool) -> None:
    if verbose:
        logging.basicConfig(level=logging.INFO)


@click.command(context_settings={"help_option_names": ["-h", "--help"]})
@click.argument("dataset_dir", type=click.Path(exists=True))
@click.argument("output_dir", type=click.Path(exists=False))
@click.argument("config_file", type=click.File(mode="r"))
@click.option(
    "-g",
    "--generate-config-file",
    help="write default config to specified file",
    type=click.File(mode="w"),
)
@click.option("--verbose", "-v", count=True, help="print progress messages")
def nrtk_perturber_cli(
    dataset_dir: str,
    output_dir: str,
    config_file: TextIO,
    generate_config_file: TextIO,
    verbose: bool,
) -> None:
    """Generate NRTK perturbed images and detections from a given set of source images and COCO-format annotations.

    Generate NRTK perturbed images and detections from a given set of source images and COCO-format annotations
    and write them to an output folder in disk. The perturbed images are stored in subfolders named after the
    chosen perturbation parameter keys and values.

    DATASET_DIR - Root directory of dataset.
    OUTPUT_DIR - Directory to write the perturbed images to.
    CONFIG_FILE - Configuration file specifying the PerturbImageFactory configuration.

    :param dataset_dir: Root directory of dataset.
    :param output_dir: Directory to write the perturbed images to.
    :param config_file: Configuration file specifying the PerturbImageFactory configuration.
    :param generate_config_file: File to write default config file, only written
        if specified.
    :param verbose: Display progress messages. Default is false.
    """
    if generate_config_file:
        config: dict[str, Any] = dict()
        config["PerturberFactory"] = make_default_config(PerturbImageFactory.get_impls())
        json.dump(config, generate_config_file, indent=4)

        exit()

    _set_logging(verbose)

    logging.info(f"Dataset path: {dataset_dir}")

    # Load COCO dataset
    coco_file = Path(dataset_dir) / "annotations.json"
    if not coco_file.is_file():
        raise ValueError("Could not identify annotations file. Expected at '[dataset_dir]/annotations.json'")
    logging.info(f"Loading kwcoco annotations from {coco_file}")
    if not is_usable:
        print("This tool requires additional dependencies, please install `nrtk-jatic[tools]`")
        exit(-1)
    kwcoco_dataset = kwcoco.CocoDataset(coco_file)

    # Load metadata, if it exists
    metadata = _load_metadata(dataset_dir, kwcoco_dataset)

    # Load config
    config = json.load(config_file)
    perturber_factory = from_config_dict(config["PerturberFactory"], PerturbImageFactory.get_impls())

    # Initialize dataset object
    input_dataset = COCOJATICObjectDetectionDataset(
        root=dataset_dir,
        kwcoco_dataset=kwcoco_dataset,
        image_metadata=metadata,
    )

    # Augment input dataset
    augmented_datasets = nrtk_perturber(maite_dataset=input_dataset, perturber_factory=perturber_factory)

    # Save each augmented dataset to its own directory
    output_path = Path(output_dir)
    img_filenames = [Path(img_path.name) for img_path in input_dataset.get_img_path_list()]
    for perturb_params, aug_dataset in augmented_datasets:
        dataset_to_coco(
            dataset=aug_dataset,
            output_dir=output_path / perturb_params,
            img_filenames=img_filenames,
            dataset_categories=input_dataset.get_categories(),
        )


if __name__ == "__main__":
    nrtk_perturber_cli()



"""
This module contains nrtk_perturber, which generates augemented MAITE dataset(s)
based on a pertrber factory configuration
"""

import itertools
import logging
from collections.abc import Iterable

import numpy as np
from maite.protocols.object_detection import Dataset
from nrtk.interfaces.perturb_image_factory import PerturbImageFactory

from nrtk_jatic.interop.object_detection.augmentation import JATICDetectionAugmentation
from nrtk_jatic.interop.object_detection.dataset import JATICObjectDetectionDataset


def nrtk_perturber(maite_dataset: Dataset, perturber_factory: PerturbImageFactory) -> Iterable[tuple[str, Dataset]]:
    """Generate augmented dataset(s) of type maite.protocols.object_detection.Dataset.

    Generate augmented dataset(s) of type maite.protocols.object_detection.Dataset
    given an input dataset of the same type and a perturber factory
    implementation. Each perturber combination will result in a newly
    generated dataset.

    :param maite_dataset: A dataset object of type maite.protocols.object_detection.Dataset
    :param perturber_factory: PerturbImageFactory implementation.
    """
    perturber_factory_config = perturber_factory.get_config()
    if "theta_keys" in perturber_factory_config:  # pyBSM doesn't follow interface rules
        perturb_factory_keys = perturber_factory_config["theta_keys"]
        thetas = perturber_factory.thetas
    else:
        perturb_factory_keys = [perturber_factory.theta_key]
        thetas = [perturber_factory.thetas]

    perturber_combinations = [dict(zip(perturb_factory_keys, v)) for v in itertools.product(*thetas)]
    logging.info(f"Perturber sweep values: {perturber_combinations}")

    # Iterate through the different perturber factory parameter combinations and
    # save the perturbed images to disk
    logging.info("Starting perturber sweep")
    augmented_datasets: list[Dataset] = []
    output_perturb_params: list[str] = []
    for i, (perturber_combo, perturber) in enumerate(zip(perturber_combinations, perturber_factory)):
        output_perturb_params.append("".join(f"_{k!s}-{v!s}" for k, v in perturber_combo.items()))

        logging.info(f"Starting perturbation for {output_perturb_params[i]}")

        aug_imgs = []
        aug_dets = []
        aug_metadata = []

        jatic_perturber = JATICDetectionAugmentation(augment=perturber)

        # Formatting data to be of batch_size=1 in order to support MAITE
        # detection protocol's expected input for Augmentation
        for idx in range(len(maite_dataset)):
            aug_img, aug_det, aug_md = jatic_perturber(
                batch=(
                    [maite_dataset[idx][0]],
                    [maite_dataset[idx][1]],
                    [maite_dataset[idx][2]],
                ),
            )
            # Appending data to separate lists in order to handle images
            # of varying sizes
            aug_imgs.append(np.asarray(aug_img)[0])
            aug_dets.append(aug_det[0])
            aug_metadata.append(aug_md[0])

        augmented_datasets.append(JATICObjectDetectionDataset(aug_imgs, aug_dets, aug_metadata))
    return zip(output_perturb_params, augmented_datasets)
